import ubelt as ub


def test_cache_stamp():
    # stamp the computation of expensive-to-compute.txt
    dpath = ub.Path.appdir('ubelt/tests', 'test-cache-stamp').ensuredir()
    ub.delete(dpath)
    ub.ensuredir(dpath)
    product = dpath / 'expensive-to-compute.txt'
    self = ub.CacheStamp('test1', dpath=dpath, depends='test1',
                         product=product, hasher=None)
    if self.expired():
        product.write_text('very expensive')
        self.renew()
    assert not self.expired()
    # corrupting the output WILL expire in non-robust mode if the size is
    # different.
    product.write_text('corrupted')
    assert self.expired()
    self.hasher = 'sha1'
    # but it will expire if we are in robust mode, even if the size is not
    # different
    assert self.expired()
    # deleting the product will cause expiration in any mode
    self.hasher = None
    ub.delete(product)
    assert self.expired()


def test_cache_stamp_corrupt_product_nohasher():
    dpath = ub.Path.appdir('ubelt/tests', 'test-cache-stamp').ensuredir()
    name = 'corrupt_product_nohasher'
    ub.delete(dpath)
    ub.ensuredir(dpath)
    product = dpath / (name + '.txt')
    self = ub.CacheStamp(name, dpath=dpath, depends=name, product=product,
                         hasher=None)
    # Disable the new (as of 1.1.0) size and mtime checks
    # note: as of version 1.1.0 we also have to disable the new size and
    # mtime checks to get a non-robust mode.
    self._expire_checks['size'] = False
    self._expire_checks['mtime'] = False
    if self.expired():
        product.write_text('very expensive')
        self.renew()
    assert not self.expired()
    # corrupting the output will not expire in non-robust mode
    product.write_text('corrupted')
    assert not self.expired()


def test_not_time_expired():
    # stamp the computation of expensive-to-compute.txt
    dpath = ub.Path.appdir('ubelt/tests', 'test-cache-stamp').ensuredir()
    ub.delete(dpath)
    ub.ensuredir(dpath)
    self = ub.CacheStamp('test1', dpath=dpath, depends='test1',
                         expires=10000, hasher=None)
    self.renew()
    assert not self.expired()


def test_time_expired():
    # stamp the computation of expensive-to-compute.txt
    dpath = ub.Path.appdir('ubelt/tests', 'test-cache-stamp').ensuredir()
    ub.delete(dpath)
    ub.ensuredir(dpath)
    self = ub.CacheStamp('test1', dpath=dpath, depends='test1',
                         expires=-10000, hasher=None)
    self.renew()
    assert self.expired() == 'expired_cert'


def test_cache_stamp_corrupt_product_hasher():
    dpath = ub.Path.appdir('ubelt/tests', 'test-cache-stamp').ensuredir()
    name = 'corrupt_product_hasher'
    ub.delete(dpath)
    ub.ensuredir(dpath)
    product = dpath / (name + '.txt')
    self = ub.CacheStamp(name, dpath=dpath, depends=name, product=product,
                         hasher='sha1')
    if self.expired():
        product.write_text('very expensive')
        self.renew()
    assert not self.expired()
    # corrupting the output will not expire in non-robust mode
    product.write_text('corrupted')
    assert self.expired()


def test_cache_stamp_multiproduct():
    import os
    # stamp the computation of expensive-to-compute.txt
    dpath = ub.Path.appdir('ubelt/tests', 'test-cache-stamp').ensuredir()
    ub.delete(dpath)
    ub.ensuredir(dpath)
    product = [
        dpath / 'product1.txt',
        os.fspath(dpath / 'product2.txt'),
        dpath / 'product3.txt',
    ]
    self = ub.CacheStamp('somedata', dpath=dpath, depends='someconfig',
                         product=product)
    if self.expired():
        for fpath in product:
            ub.Path(fpath).write_text('very expensive')
        self.renew()
    assert not self.expired()
    ub.Path(product[1]).write_text('corrupted')
    assert self.expired()


def test_cache_stamp_noproduct():
    # stamp the computation of expensive-to-compute.txt
    dpath = ub.Path.appdir('ubelt/tests', 'test-cache-stamp').ensuredir()
    ub.delete(dpath)
    ub.ensuredir(dpath)
    name = 'noproduct'
    product = dpath / (name + '.txt')
    self = ub.CacheStamp('somedata', dpath=dpath, depends='someconfig', product=None)
    if self.expired():
        product.write_text('very expensive')
        self.renew()
    assert not self.expired()
    product.write_text('corrupted')
    assert not self.expired()



import io
from typing import List
from typing import Type
from types import TracebackType
from _typeshed import Incomplete


class TeeStringIO(io.StringIO):
    redirect: io.IOBase | None
    buffer: Incomplete

    def __init__(self, redirect: io.IOBase | None = None) -> None:
        ...

    def isatty(self) -> bool:
        ...

    def fileno(self) -> int:
        ...

    @property
    def encoding(self):
        ...

    @encoding.setter
    def encoding(self, value) -> None:
        ...

    def write(self, msg: str):
        ...

    def flush(self):
        ...


class CaptureStream:
    ...


class CaptureStdout(CaptureStream):
    text: str | None
    parts: List[str]
    cap_stdout: None | TeeStringIO
    orig_stdout: io.TextIOBase
    started: bool
    enabled: bool
    suppress: bool

    def __init__(self, suppress: bool = True, enabled: bool = True) -> None:
        ...

    def log_part(self) -> None:
        ...

    def start(self) -> None:
        ...

    def stop(self) -> None:
        ...

    def __enter__(self):
        ...

    def __del__(self) -> None:
        ...

    def close(self) -> None:
        ...

    def __exit__(self, ex_type: Type[BaseException] | None,
                 ex_value: BaseException | None,
                 ex_traceback: TracebackType | None) -> bool | None:
        ...



"""
Warning:

    This module is deprecated. Use :mod:`ubelt.util_repr` instead.
"""
from .util_repr import urepr, ReprExtensions, _REPR_EXTENSIONS


def repr2(data, **kwargs):
    """
    Alias of :func:`ubelt.util_repr.urepr`.

    Warning:
        Deprecated for urepr

    Example:
        >>> # Test that repr2 remains backwards compatible
        >>> import ubelt as ub
        >>> dict_ = {
        ...     'custom_types': [slice(0, 1, None), 1/3],
        ...     'nest_dict': {'k1': [1, 2, {3: {4, 5}}],
        ...                   'key2': [1, 2, {3: {4, 5}}],
        ...                   'key3': [1, 2, {3: {4, 5}}],
        ...                   },
        ...     'nest_dict2': {'k': [1, 2, {3: {4, 5}}]},
        ...     'nested_tuples': [tuple([1]), tuple([2, 3]), frozenset([4, 5, 6])],
        ...     'one_tup': tuple([1]),
        ...     'simple_dict': {'spam': 'eggs', 'ham': 'jam'},
        ...     'simple_list': [1, 2, 'red', 'blue'],
        ...     'odict': ub.odict([(2, '1'), (1, '2')]),
        ... }
        >>> import pytest
        >>> with pytest.warns(DeprecationWarning):
        >>>     result = ub.repr2(dict_, nl=1, precision=2)
        >>> print(result)
        {
            'custom_types': [slice(0, 1, None), 0.33],
            'nest_dict': {'k1': [1, 2, {3: {4, 5}}], 'key2': [1, 2, {3: {4, 5}}], 'key3': [1, 2, {3: {4, 5}}]},
            'nest_dict2': {'k': [1, 2, {3: {4, 5}}]},
            'nested_tuples': [(1,), (2, 3), {4, 5, 6}],
            'odict': {2: '1', 1: '2'},
            'one_tup': (1,),
            'simple_dict': {'ham': 'jam', 'spam': 'eggs'},
            'simple_list': [1, 2, 'red', 'blue'],
        }
    """
    from ubelt.util_deprecate import schedule_deprecation
    schedule_deprecation(
        modname='ubelt', name='repr2', type='function',
        migration='use urepr instead',
        deprecate='1.2.5', error='2.0.0', remove='2.1.0',
    )
    kwargs['_dict_sort_behavior'] = kwargs.get('_dict_sort_behavior', 'old')
    return urepr(data, **kwargs)


repr2.extensions = urepr.extensions
repr2.register = urepr.register


# Deprecated aliases
FormatterExtensions = ReprExtensions
_FORMATTER_EXTENSIONS = _REPR_EXTENSIONS


__all__ = ['repr2', 'urepr', 'FormatterExtensions']



from os import PathLike
import concurrent
import concurrent.futures
from typing import Iterable


class DownloadManager:
    download_root: str | PathLike
    cache: bool

    def __init__(self,
                 download_root: str | PathLike | None = None,
                 mode: str = 'thread',
                 max_workers: int | None = None,
                 cache: bool = True) -> None:
        ...

    def submit(self,
               url: str | PathLike,
               dst: str | None = None,
               hash_prefix: str | None = None,
               hasher: str = 'sha256') -> concurrent.futures.Future:
        ...

    def as_completed(self,
                     prog: None | bool | type = None,
                     desc: str | None = None,
                     verbose: int = 1):
        ...

    def shutdown(self) -> None:
        ...

    def __iter__(self) -> Iterable:
        ...

    def __len__(self) -> int:
        ...



import ubelt as ub
from os.path import expanduser, basename


def test_compressuser_without_home():
    username = basename(expanduser('~'))
    not_the_user = 'foobar_' + username
    ub.shrinkuser(not_the_user) == not_the_user


def test_find_path_no_path():
    candidates = list(ub.find_path('does-not-exist', path=[]))
    assert len(candidates) == 0


def _available_prog():
    # Try and find a program that exists on the machine
    import pytest
    common_progs = ['ls', 'ping', 'which']
    prog_name = None
    for cand_prog_name in common_progs:
        if ub.find_exe(cand_prog_name):
            prog_name = cand_prog_name
            break
    else:
        pytest.skip((
            'Common progs {} are not installed. '
            'Are we on a weird machine?').format(common_progs))
    return prog_name


def test_find_exe_idempotence():
    prog_name = _available_prog()
    prog_fpath = ub.find_exe(prog_name)
    assert prog_fpath == ub.find_exe(prog_fpath), (
        'find_exe with an existing path should work')


def test_find_exe_no_exist():
    assert ub.find_exe('!noexist', multi=False) is None, (
        'multi=False not found should return None')
    assert ub.find_exe('!noexist', multi=True) == [], (
        'multi=True not found should return an empty list')


if __name__ == '__main__':
    """
        pytest ubelt/tests/test_platform.py
    """

    import xdoctest
    xdoctest.doctest_module(__file__)



import ubelt as ub
import os
import pytest
import sys
from os.path import basename, join, exists
import platform


IS_PYPY = platform.python_implementation() == 'PyPy'
IS_WIN32 = sys.platform.startswith('win32')


TIMEOUT = (15 if IS_PYPY else 5) * 30


@pytest.mark.timeout(TIMEOUT)
def test_download_no_fpath():
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.Path.appdir('ubelt/tests/test_download').ensuredir()
    fname = basename(url)
    fpath = join(dpath, fname)

    ub.delete(fpath)
    assert not exists(fpath)

    got_fpath = ub.download(url, appname='ubelt/tests/test_download')

    assert got_fpath == fpath
    assert exists(fpath)


@pytest.mark.timeout(TIMEOUT)
def test_download_with_fpath():
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url(1201)

    dpath = ub.Path.appdir('ubelt/tests/test_download').ensuredir()
    fname = basename(url)
    fpath = join(dpath, fname)

    ub.delete(fpath)
    assert not exists(fpath)

    got_fpath = ub.download(url, fpath=fpath,
                            appname='ubelt/tests/test_download')
    assert got_fpath == fpath
    assert exists(fpath)

    with open(got_fpath, 'rb') as file:
        data = file.read()
    assert len(data) > 1200, 'should have downloaded some bytes'


@pytest.mark.timeout(TIMEOUT)
def test_download_chunksize():
    # url = 'https://www.dropbox.com/s/jl506apezj42zjz/ibeis-win32-setup-ymd_hm-2015-08-01_16-28.exe?dl=1'
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.Path.appdir('ubelt/tests/test_download').ensuredir()
    fname = basename(url)
    fpath = join(dpath, fname)

    ub.delete(fpath)
    assert not exists(fpath)

    got_fpath = ub.download(url, chunksize=2, appname='ubelt/tests/test_download')

    assert got_fpath == fpath
    assert exists(fpath)


@pytest.mark.timeout(TIMEOUT)
def test_download_cover_hashers():
    # url = 'https://www.dropbox.com/s/jl506apezj42zjz/ibeis-win32-setup-ymd_hm-2015-08-01_16-28.exe?dl=1'
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.Path.appdir('ubelt/tests/test_download').ensuredir()
    fname = basename(url)

    # add coverage for different hashers
    ub.download(url, hasher='md5', hash_prefix='e09c80c42fda55f9d992e59ca6b33',
                dpath=dpath, fname=fname)
    ub.download(url, hasher='sha256', hash_prefix='bf2cb58a68f684d95a3b78ef8f',
                dpath=dpath, fname=fname)


@pytest.mark.timeout(TIMEOUT)
def test_download_hashalgo():
    # url = 'https://www.dropbox.com/s/jl506apezj42zjz/ibeis-win32-setup-ymd_hm-2015-08-01_16-28.exe?dl=1'
    import hashlib

    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.Path.appdir('ubelt/tests/test_download').ensuredir()
    fname = basename(url)
    fpath = join(dpath, fname)

    ub.delete(fpath)
    assert not exists(fpath)

    got_fpath = ub.download(url,
                            hash_prefix='e09c80c42fda55f9d992e59ca6b3307d',
                            appname='ubelt/tests/test_download',
                            hasher=hashlib.md5())

    assert got_fpath == fpath
    assert exists(fpath)


@pytest.mark.timeout(TIMEOUT)
def test_grabdata_cache():
    """
    Check where the url is downloaded to when fpath is not specified.
    """
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.Path.appdir('ubelt/tests/test_download').ensuredir()
    fname = basename(url)
    fpath = join(dpath, fname)

    got_fpath = ub.grabdata(url, appname='ubelt/tests/test_download')
    assert got_fpath == fpath
    assert exists(fpath)

    ub.delete(fpath)
    assert not exists(fpath)

    ub.grabdata(url, appname='ubelt/tests/test_download')
    assert exists(fpath)


@pytest.mark.timeout(TIMEOUT)
def test_grabdata_nohash():
    """
    Check where the url is downloaded to when fpath is not specified.
    """
    url = _demo_url()
    dpath = ub.Path.appdir('ubelt/tests/test_download/test-grabdata-nohash').ensuredir()
    fname = basename(url)
    fpath = (dpath / fname).delete()
    assert not fpath.exists()
    ub.grabdata(url, fpath=fpath, hasher=None, verbose=10)
    assert fpath.exists()
    # Even without the hasher, if the size of the data changes at all
    # we should be able to detect and correct it.
    orig_text = fpath.read_text()
    fpath.write_text('corrupted')
    ub.grabdata(url, fpath=fpath, hasher=None, verbose=10)
    assert fpath.read_text() == orig_text


@pytest.mark.timeout(TIMEOUT)
def test_grabdata_url_only():
    """
    Check where the url is downloaded to when fpath is not specified.
    """
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.Path.appdir('ubelt')
    fname = basename(url)
    fpath = os.fspath(dpath / fname)

    got_fpath = ub.grabdata(url)
    assert got_fpath == fpath
    assert exists(fpath)


@pytest.mark.timeout(TIMEOUT)
def test_grabdata_with_fpath():
    """
    Check where the url is downloaded to when fpath is not specified.
    """
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.Path.appdir('ubelt/tests/test_download').ensuredir()
    fname = basename(url)
    fpath = join(dpath, fname)

    got_fpath = ub.grabdata(url, fpath=fpath, verbose=3)
    assert got_fpath == fpath
    assert exists(fpath)

    ub.delete(fpath)
    assert not exists(fpath)

    ub.grabdata(url, fpath=fpath, verbose=3)
    assert exists(fpath)


def test_grabdata_value_error():
    """
    Check where the url is downloaded to when fpath is not specified.
    """
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.Path.appdir('ubelt/tests/test_download').ensuredir()
    fname = basename(url)
    fpath = join(dpath, fname)

    with pytest.raises(ValueError):
        ub.grabdata(url, fname=fname, fpath=fpath, dpath=dpath)

    with pytest.raises(ValueError):
        ub.grabdata(url, fname=fname, fpath=fpath)

    with pytest.raises(ValueError):
        ub.grabdata(url, dpath=dpath, fpath=fpath)

    with pytest.raises(ValueError):
        ub.grabdata(url, fpath=fpath, appname='foobar')

    with pytest.raises(ValueError):
        ub.grabdata(url, dpath=dpath, appname='foobar')


@pytest.mark.timeout(TIMEOUT * 2)
def test_download_bad_url():
    """
    Check that we error when the url is bad

    Notes:
        For some reason this can take a long time to realize there is no URL,
        even if the timeout is specified and fairly low.

    CommandLine:
        python tests/test_download.py test_download_bad_url --verbose
    """
    import pytest
    pytest.skip('This takes a long time to timeout and I dont understand why')

    url = 'http://www.a-very-incorrect-url.gov/does_not_exist.txt'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')

    # Ensure the opener exist
    # import urllib.request as urllib_x
    from urllib.error import URLError  # NOQA
    # if urllib_x._opener is None:
    #     urllib_x.install_opener(urllib_x.build_opener())

    dpath = ub.Path.appdir('ubelt/tests/test_download').ensuredir()
    fname = basename(url)
    fpath = join(dpath, fname)

    ub.delete(fpath)
    assert not exists(fpath)

    with pytest.raises(URLError):
        ub.download(url, fpath=fpath, verbose=1, timeout=1.0)


@pytest.mark.timeout(TIMEOUT)
def test_grabdata_fname_only():
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    # fname = 'mario.png'

    url = _demo_url()

    dpath = ub.Path.appdir('ubelt')
    fname = 'custom_text.txt'
    fpath = os.fspath(dpath / fname)

    got_fpath = ub.grabdata(url, fname=fname)
    assert got_fpath == fpath
    assert exists(fpath)


@pytest.mark.timeout(TIMEOUT)
def test_grabdata_dpath_only():
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    dpath = ub.Path.appdir('ubelt/tests/test_download').ensuredir()
    fname = basename(url)
    fpath = join(dpath, fname)

    got_fpath = ub.grabdata(url, dpath=dpath)
    assert got_fpath == fpath
    assert exists(fpath)


@pytest.mark.timeout(TIMEOUT)
def test_grabdata_fpath_and_dpath():
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')
    url = _demo_url()

    with pytest.raises(ValueError):
        ub.grabdata(url, fpath='foo', dpath='bar')


# @pytest.mark.timeout(TIMEOUT)
def test_grabdata_hash_typo():
    """
    CommandLine:
        pytest ~/code/ubelt/tests/test_download.py -k test_grabdata_hash_typo --network -s
        xdoctest ~/code/ubelt/tests/test_download.py test_grabdata_hash_typo --network

    """
    # url = 'https://www.dropbox.com/s/jl506apezj42zjz/ibeis-win32-setup-ymd_hm-2015-08-01_16-28.exe?dl=1'
    # url = 'http://i.imgur.com/rqwaDag.png'
    # if not ub.argflag('--network'):
    #     pytest.skip('not running network tests')

    url = _demo_url()

    dpath = ub.Path.appdir('ubelt/tests/test_download')
    fname = basename(url)
    fpath = dpath / fname
    stamp_fpath = fpath.augment(tail='.stamp_md5.json')

    for verbose in [5]:
        fpath.delete()
        stamp_fpath.delete()
        assert not exists(fpath)

        print('[STEP1] Downloading file, but we have a typo in the hash')
        with pytest.raises(RuntimeError):
            got_fpath = ub.grabdata(
                url, hash_prefix='e09c80c42fda5-typo-5f9d992e59ca6b3307d',
                hasher='md5', verbose=verbose,
                appname='ubelt/tests/test_download')
        assert fpath.exists()
        real_hash = ub.hash_file(fpath, hasher='md5')
        real_hash

        print('[STEP2] Fixing the typo recomputes the hash, but does not redownload the file')
        got_fpath = ub.grabdata(url,
                                hash_prefix='e09c80c42fda55f9d992e59ca6b3307d',
                                hasher='md5', verbose=verbose,
                                appname='ubelt/tests/test_download')
        assert ub.Path(got_fpath).resolve() == fpath.resolve()
        assert fpath.exists()

        # If we delete the .hash file we will simply recompute
        stamp_fpath.delete()
        print('[STEP3] Deleting the hash file recomputes the hash')
        got_fpath = ub.grabdata(url, fpath=fpath,
                                hash_prefix='e09c80c42fda55f9d992e59ca6b3307d',
                                hasher='md5', verbose=verbose)
        assert stamp_fpath.exists()


def test_deprecated_grabdata_args():
    # with pytest.warns(DeprecationWarning):
    with pytest.raises(RuntimeError):
        import hashlib
        url = _demo_url()
        # dpath = ub.Path.appdir('ubelt/tests/test_download').ensuredir()
        # fname = basename(url)
        # fpath = join(dpath, fname)
        got_fpath = ub.grabdata(
            url, hash_prefix='e09c80c42fda55f9d992e59ca6b3307d',
            hasher=hashlib.md5())
        got_fpath


def _devcheck_progres_download_bar():
    """
    import sys, ubelt
    sys.path.append(ubelt.expandpath('~/remote/toothbrush/code/ubelt/tests'))
    from test_download import SingletonTestServer
    import ubelt as ub
    """
    self = SingletonTestServer.instance()
    urls = self.write_file(filebytes=85717624)

    import time

    class DummyIO:
        def write(self, msg):
            time.sleep(0.0005)
            ...

    file = DummyIO()
    url = urls[0]
    dl_file = ub.download(url, fpath=file, progkw=dict(desc='dling'))
    dl_file


class SingletonTestServer(ub.NiceRepr):
    """
    A singleton class used for testing.

    This could be done via a pytest fixture, but... I don't want to use
    fixtures until its easy and clear how to make an instance of them in an
    independent IPython session.

    CommandLine:
        xdoctest -m tests/test_download.py SingletonTestServer

    Note:
        We rely on python process close mechanisms to clean this server up.
        Might need to re-investigate this in the future.

    Ignore:
        >>> self = SingletonTestServer.instance()
        >>> print('self = {!r}'.format(self))
        >>> url = self.urls[0]
        >>> print('url = {!r}'.format(url))
        >>> dl_file = ub.download(url)
    """

    _instance = None

    @classmethod
    def instance(cls):
        if cls._instance is not None:
            self = cls._instance
        else:
            self = cls()
            cls._instance = self
        return self

    def close(self):
        if self.proc.poll() is None:
            self.proc.terminate()
            self.proc.wait()
        self.__class__.instance = None

    def __nice__(self):
        return '{} - {}'.format(self.root_url, self.proc.returncode)

    def __init__(self):
        import requests
        import time
        import sys
        import ubelt as ub
        import socket
        from contextlib import closing
        def find_free_port():
            """
            References:
                .. [SO1365265] https://stackoverflow.com/questions/1365265/on-localhost-how-do-i-pick-a-free-port-number
            """
            with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
                s.bind(('', 0))
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                return s.getsockname()[1]

        # Find an open port
        port = find_free_port()
        print('port = {!r}'.format(port))

        dpath = ub.Path.appdir('ubelt/tests/test_download/simple_server').ensuredir()

        if sys.platform.startswith('win32'):
            pyexe = 'python'
        else:
            pyexe = sys.executable

        server_cmd = [
            pyexe, '-m', 'http.server', str(port)
        ]
        info = ub.cmd(server_cmd, detach=True, cwd=dpath)
        proc = info['proc']
        self.proc = proc
        self.dpath = dpath
        self.root_url = 'http://localhost:{}'.format(port)

        if IS_PYPY and IS_WIN32:
            # not sure why
            import pytest
            pytest.skip('not sure why download tests are failing on pypy win32')
            init_sleeptime = 0.5
            fail_sleeptime = 0.3
            timeout = 10
        else:
            init_sleeptime = 0.002
            fail_sleeptime = 0.01
            timeout = 1

        time.sleep(init_sleeptime)
        # Wait for the server to be alive
        status_code = None
        max_tries = 300
        for _ in range(max_tries):
            try:
                resp = requests.get(self.root_url, timeout=timeout)
            except requests.exceptions.ConnectionError:
                time.sleep(fail_sleeptime)
            else:
                status_code = resp.status_code
            if status_code == 200:
                break

        poll_ret = self.proc.poll()

        if poll_ret is not None:
            print('poll_ret = {!r}'.format(poll_ret))
            print(self.proc.communicate())
            raise AssertionError('Simple server did not start {}'.format(poll_ret))

        self.urls = []
        self.write_file()

    def write_file(self, filebytes=10, num_files=1):
        fnames = ['file_{}_{}.txt'.format(filebytes, i) for i in range(num_files)]
        for fname in fnames:
            # data = ''.join(random.choices(string.ascii_letters, k=filebytes))
            data = 'a' * filebytes
            fpath = join(self.dpath, fname)
            with open(fpath, 'w') as file:
                file.write(data)
        urls = ['{}/{}'.format(self.root_url, fname) for fname in fnames]
        self.urls.extend(urls)
        return urls


def test_local_download():
    server = SingletonTestServer.instance()
    url = server.write_file(filebytes=int(10 * 2 ** 20))[0]
    # also test with a timeout for lazy coverage
    ub.download(url, timeout=3000)


def _demo_url(num_bytes=None):
    REAL_URL = False
    if REAL_URL:
        url = 'http://i.imgur.com/rqwaDag.png'
        if not ub.argflag('--network'):
            pytest.skip('not running network tests')
    else:
        if num_bytes is None:
            url = SingletonTestServer.instance().urls[0]
        else:
            url = SingletonTestServer.instance().write_file(num_bytes)[0]
    return url


@pytest.mark.timeout(TIMEOUT)
def test_download_with_progkw():
    """
    Test that progkw is properly passed through to ub.download
    """
    url = _demo_url(128 * 10)
    dpath = ub.Path.appdir('ubelt/tests/test_download').ensuredir()
    fname = basename(url)
    fpath = join(dpath, fname)
    with ub.CaptureStdout() as cap:
        ub.download(url, fpath=fpath, progkw={'verbose': 3, 'freq': 1, 'adjust': False, 'time_thresh': 0}, chunksize=128)
    assert len(cap.text.split('\n')) > 10


@pytest.mark.timeout(TIMEOUT)
def test_download_with_filesize():
    """
    Test that progkw is properly passed through to ub.download
    """
    url = _demo_url(128 * 10)
    dpath = ub.Path.appdir('ubelt/tests/test_download').ensuredir()
    fname = basename(url)
    fpath = join(dpath, fname)
    with ub.CaptureStdout() as cap:
        ub.download(url, filesize=11, fpath=fpath, progkw={'verbose': 3, 'freq': 1, 'adjust': False, 'time_thresh': 0}, chunksize=128)
    import re
    assert re.search(r'\d\d\d\d\.\d\d%', cap.text), 'should report over 100%'


def make_stat_dict(stat_obj):
    # Convert the stat tuple to a dict we can manipulate
    # and ignore access time
    ignore_keys = {'st_atime', 'st_atime_ns'}
    return {
        k: getattr(stat_obj, k) for k in dir(stat_obj)
        if k.startswith('st_') and k not in ignore_keys}


def test_grabdata():
    import ubelt as ub
    import json
    import time
    # fname = 'foo.bar'
    # url = 'http://i.imgur.com/rqwaDag.png'
    # prefix1 = '944389a39dfb8fa9'
    url = _demo_url(128 * 11)
    prefix1 = 'b7fa848cd088ae842a89'
    fname = 'foo2.bar'
    #
    print('1. Download the file once')
    fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1, hasher='sha512')
    stat0 = make_stat_dict(ub.Path(fpath).stat())
    stamp_fpath = ub.Path(fpath).augment(tail='.stamp_sha512.json')
    assert json.loads(stamp_fpath.read_text())['hash'][0].startswith(prefix1)
    #
    print("2. Rerun and check that the download doesn't happen again")
    fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1)
    stat1 = make_stat_dict(ub.Path(fpath).stat())
    assert stat0 == stat1, 'the file should not be modified'
    #
    print('3. Set redo=True, which should force a redownload')
    sleep_time = 0.1
    num_tries = 60
    for _ in range(num_tries):
        fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1, redo=True,
                            hasher='sha512')
        stat2 = make_stat_dict(ub.Path(fpath).stat())
        # Note: the precision of mtime is too low for this test work reliably
        # https://apenwarr.ca/log/20181113
        if stat2 != stat1:
            break
        print('... Sometimes the redownload happens so fast we need to '
              'wait to notice the file is actually different')
        time.sleep(sleep_time)
    else:
        raise AssertionError(
            'the file stat should be modified, we waited over {}s.'.format(
                sleep_time * num_tries))
    #
    print('4. Check that a redownload occurs when the stamp is changed')
    stamp_fpath.write_text('corrupt-stamp')
    fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1, hasher='sha512')
    assert json.loads(stamp_fpath.read_text())['hash'][0].startswith(prefix1)
    #
    print('5. Check that a redownload occurs when the stamp is removed')
    ub.delete(stamp_fpath)
    fpath = ub.Path(fpath)
    fpath.write_text('corrupt-stamp')
    assert not ub.hash_file(fpath, base='hex', hasher='sha512').startswith(prefix1)
    fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1, hasher='sha512')
    assert ub.hash_file(fpath, base='hex', hasher='sha512').startswith(prefix1)


def test_grabdata_same_fpath_different_url():
    url1 = _demo_url(128 * 11)
    url2 = _demo_url(128 * 12)
    url3 = _demo_url(128 * 13)

    fname = 'foobar'
    fpath1 = ub.grabdata(url1, fname=fname, hash_prefix='b7fa848cd088ae842a89ef', hasher='sha512', verbose=100)
    stat1 = make_stat_dict(ub.Path(fpath1).stat())

    # Should requesting a new url, even with the same fpath, cause redownload?
    fpath2 = ub.grabdata(url2, fname=fname, hash_prefix=None, hasher='sha512', verbose=100)
    stat2 = make_stat_dict(ub.Path(fpath2).stat())

    fpath3 = ub.grabdata(url3, fname=fname, hash_prefix=None, hasher='sha512', verbose=100)
    stat3 = make_stat_dict(ub.Path(fpath3).stat())

    assert stat1 != stat2, 'the stats will change because we did not specify a hash prefix'
    assert stat2 == stat3, 'we may change this behavior in the future'

    fpath3 = ub.grabdata(url2, fname=fname, hash_prefix='43f92597d7eb08b57c88b6', hasher='sha512', verbose=100)
    stat3 = make_stat_dict(ub.Path(fpath3).stat())
    assert stat1 != stat3, 'if we do specify a new hash, we should get a new download'
    assert url1 != url2, 'urls should be different'
    assert ub.allsame([fpath1, fpath2, fpath3]), 'all fpaths should be the same'


def test_grabdata_delete_hash_stamp():
    import ubelt as ub
    fname = 'foo3.bar'
    url = _demo_url(128 * 12)
    prefix1 = '43f92597d7eb08b57c88b636'
    fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1)
    stamp_fpath = ub.Path(fpath + '.stamp_sha512.json')
    ub.delete(stamp_fpath)
    fpath = ub.grabdata(url, fname=fname, hash_prefix=prefix1)


def test_download_with_io():
    import ubelt as ub
    import io
    url = _demo_url(128 * 3)
    file = io.BytesIO()
    fpath = ub.download(url, file)
    assert fpath is file
    file.seek(0)
    data = file.read()
    hashstr = ub.hash_data(data, hasher='sha1')
    assert hashstr.startswith('45a5c851bf12d1')


def test_download_with_sha1_hasher():
    import ubelt as ub
    url = _demo_url(128 * 4)
    ub.download(url, hasher='sha1', hash_prefix='164557facb7392')


def setup_module(module):
    """ setup any state specific to the execution of the given module."""
    SingletonTestServer.instance()


def teardown_module(module):
    """ teardown any state that was previously setup with a setup_module
    method.
    """
    SingletonTestServer.instance().close()


if __name__ == '__main__':
    """
    CommandLine:
        pytest ubelt/tests/test_download.py
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
UBelt is a "utility belt" of commonly needed utility and helper functions.  It
is a currated collection of top-level utilities with functionality that falls
into a mixture of categories.

The source code is available at `https://github.com/Erotemic/ubelt <https://github.com/Erotemic/ubelt>`_.
We also have `Jupyter notebook demos <https://github.com/Erotemic/ubelt/blob/master/docs/notebooks/Ubelt%20Demo.ipynb>`_.

The ubelt API is organized by submodules containing related functionality.
Each submodule contains top level overview documentation, and each function
contains a docstring with at least one example.

NOTE: The `README <https://github.com/Erotemic/ubelt#readme>`_ on github
contains information and examples complementary to these docs.
"""

__dev__ = """
AutogenInit:
    mkinit ubelt --diff
    mkinit ubelt -w  # todo: get sphinx to ignore this
    # TODO: Lazy imports with mkinit (requires python 3.7)

Testing:
    xdoctest ubelt
"""

__version__ = '1.3.5'

# Deprecated functions
from ubelt.util_platform import (
    ensure_app_cache_dir, ensure_app_config_dir,
    ensure_app_data_dir, get_app_cache_dir, get_app_config_dir,
    get_app_data_dir,
)

from ubelt.util_io import (readfrom, writeto,)
from ubelt.util_str import (ensure_unicode,)
from ubelt import util_format
from ubelt.util_format import FormatterExtensions, repr2


__ignore__ = [
    'ensure_app_cache_dir',
    'ensure_app_config_dir',
    'ensure_app_data_dir',
    'get_app_cache_dir',
    'get_app_config_dir',
    'get_app_data_dir',
    'readfrom',
    'writeto',
    'ensure_unicode',
]

__explicit__ = [
    'ensure_app_cache_dir',
    'ensure_app_config_dir',
    'ensure_app_data_dir',
    'get_app_cache_dir',
    'get_app_config_dir',
    'get_app_data_dir',
    'readfrom',
    'writeto',
    'ensure_unicode',
    'util_format',
    'repr2',
    'FormatterExtensions',
]


__submodules__ = {
    'util_arg': None,
    'util_cache': None,
    'util_colors': None,
    'util_const': None,
    'util_cmd': None,
    'util_dict': None,
    'util_deprecate': None,
    'util_download': None,
    'util_download_manager': None,
    'util_func': None,
    'util_repr': None,
    'util_futures': None,
    'util_io': None,
    'util_links': None,
    'util_list': None,
    'util_hash': None,
    'util_import': None,
    'util_indexable': None,
    'util_memoize': None,
    'util_mixins': None,
    'util_path': None,
    'util_platform': None,
    'util_str': None,
    'util_stream': None,
    'util_time': None,
    'util_zip': None,
    'orderedset': None,
    'progiter': None,
}

from ubelt import orderedset
from ubelt import progiter
from ubelt import util_arg
from ubelt import util_cache
from ubelt import util_cmd
from ubelt import util_colors
from ubelt import util_const
from ubelt import util_deprecate
from ubelt import util_dict
from ubelt import util_download
from ubelt import util_download_manager
from ubelt import util_func
from ubelt import util_futures
from ubelt import util_hash
from ubelt import util_import
from ubelt import util_indexable
from ubelt import util_io
from ubelt import util_links
from ubelt import util_list
from ubelt import util_memoize
from ubelt import util_mixins
from ubelt import util_path
from ubelt import util_platform
from ubelt import util_repr
from ubelt import util_str
from ubelt import util_stream
from ubelt import util_time
from ubelt import util_zip


# Deprecated parts of the top-level API
# These functions are mostly moved into internal classes
__deprecated__ = [
    'AutoOrderedDict',
    'dict_diff',
    'dict_isect',
    'dict_subset',
    'invert_dict',
    'map_keys',
    'map_vals',
    'map_values',
    'sorted_keys',
    'sorted_vals',
    'delete',
    'touch',
    'augpath',
    'ensuredir',
    'expandpath',
    'shrinkuser',
    'userhome',
]


from ubelt.util_arg import (argflag, argval,)
from ubelt.util_cache import (CacheStamp, Cacher,)
from ubelt.util_colors import (NO_COLOR, color_text, highlight_code,)
from ubelt.util_const import (NoParam,)
from ubelt.util_cmd import (cmd,)
from ubelt.util_dict import (AutoDict, AutoOrderedDict, SetDict, UDict, ddict,
                             dict_diff, dict_hist, dict_isect, dict_subset,
                             dict_union, dzip, find_duplicates, group_items,
                             invert_dict, map_keys, map_vals, map_values,
                             named_product, odict, sdict, sorted_keys,
                             sorted_vals, sorted_values, udict, varied_values,)
from ubelt.util_deprecate import (schedule_deprecation,)
from ubelt.util_download import (download, grabdata,)
from ubelt.util_download_manager import (DownloadManager,)
from ubelt.util_func import (compatible, identity, inject_method,)
from ubelt.util_repr import (ReprExtensions, urepr,)
from ubelt.util_futures import (Executor, JobPool,)
from ubelt.util_io import (delete, touch,)
from ubelt.util_links import (symlink,)
from ubelt.util_list import (allsame, argmax, argmin, argsort, argunique,
                             boolmask, chunks, compress, flatten, iter_window,
                             iterable, peek, take, unique, unique_flags,)
from ubelt.util_hash import (hash_data, hash_file,)
from ubelt.util_import import (import_module_from_name,
                               import_module_from_path, modname_to_modpath,
                               modpath_to_modname, split_modpath,)
from ubelt.util_indexable import (IndexableWalker, indexable_allclose,)
from ubelt.util_memoize import (memoize, memoize_method, memoize_property,)
from ubelt.util_mixins import (NiceRepr,)
from ubelt.util_path import (ChDir, Path, TempDir, augpath, ensuredir,
                             expandpath, shrinkuser, userhome,)
from ubelt.util_platform import (DARWIN, LINUX, POSIX, WIN32, find_exe,
                                 find_path, platform_cache_dir,
                                 platform_config_dir, platform_data_dir,)
from ubelt.util_str import (codeblock, hzcat, indent, paragraph,)
from ubelt.util_stream import (CaptureStdout, CaptureStream, TeeStringIO,)
from ubelt.util_time import (Timer, timeparse, timestamp,)
from ubelt.util_zip import (split_archive, zopen,)
from ubelt.orderedset import (OrderedSet, oset,)
from ubelt.progiter import (ProgIter,)

__all__ = ['AutoDict', 'AutoOrderedDict', 'CacheStamp', 'Cacher',
           'CaptureStdout', 'CaptureStream', 'ChDir', 'DARWIN',
           'DownloadManager', 'Executor', 'FormatterExtensions',
           'IndexableWalker', 'JobPool', 'LINUX', 'NO_COLOR', 'NiceRepr',
           'NoParam', 'OrderedSet', 'POSIX', 'Path', 'ProgIter',
           'ReprExtensions', 'SetDict', 'TeeStringIO', 'TempDir', 'Timer',
           'UDict', 'WIN32', 'allsame', 'argflag', 'argmax', 'argmin',
           'argsort', 'argunique', 'argval', 'augpath', 'boolmask', 'chunks',
           'cmd', 'codeblock', 'color_text', 'compatible', 'compress', 'ddict',
           'delete', 'dict_diff', 'dict_hist', 'dict_isect', 'dict_subset',
           'dict_union', 'download', 'dzip', 'ensure_app_cache_dir',
           'ensure_app_config_dir', 'ensure_app_data_dir', 'ensure_unicode',
           'ensuredir', 'expandpath', 'find_duplicates', 'find_exe',
           'find_path', 'flatten', 'get_app_cache_dir', 'get_app_config_dir',
           'get_app_data_dir', 'grabdata', 'group_items', 'hash_data',
           'hash_file', 'highlight_code', 'hzcat', 'identity',
           'import_module_from_name', 'import_module_from_path', 'indent',
           'indexable_allclose', 'inject_method', 'invert_dict', 'iter_window',
           'iterable', 'map_keys', 'map_vals', 'map_values', 'memoize',
           'memoize_method', 'memoize_property', 'modname_to_modpath',
           'modpath_to_modname', 'named_product', 'odict', 'orderedset',
           'oset', 'paragraph', 'peek', 'platform_cache_dir',
           'platform_config_dir', 'platform_data_dir', 'progiter', 'readfrom',
           'repr2', 'schedule_deprecation', 'sdict', 'shrinkuser',
           'sorted_keys', 'sorted_vals', 'sorted_values', 'split_archive',
           'split_modpath', 'symlink', 'take', 'timeparse', 'timestamp',
           'touch', 'udict', 'unique', 'unique_flags', 'urepr', 'userhome',
           'util_arg', 'util_cache', 'util_cmd', 'util_colors', 'util_const',
           'util_deprecate', 'util_dict', 'util_download',
           'util_download_manager', 'util_format', 'util_func', 'util_futures',
           'util_hash', 'util_import', 'util_indexable', 'util_io',
           'util_links', 'util_list', 'util_memoize', 'util_mixins',
           'util_path', 'util_platform', 'util_repr', 'util_str',
           'util_stream', 'util_time', 'util_zip', 'varied_values', 'writeto',
           'zopen']



"""
This is a specialized set of tests for the util_import module on editable
installs, specifically with the new setuptools editable hooks (v64.0.0).
https://setuptools.pypa.io/en/latest/userguide/development_mode.html
https://setuptools.pypa.io/en/latest/history.html#id37

Running the setup and teardown for this test is very expensive wrt how long
this test takes versus others in this library.
We should look into if there is a cheaper way to emulate it.
What we could do is run the expensive test once, and serialize the outputs it
produces so we can simply reconstruct the environment.
"""
import os
import sys


class ProjectStructure():
    """
    Method to help setup and teardown a demo package installed in editable
    mode.

    Ignore:
        import ubelt as ub
        import sys, ubelt
        import xdev
        sys.path.append(ubelt.expandpath('~/code/ubelt/tests'))
        from test_editable_modules import *  # NOQA
        dpath = ub.Path.appdir('ubelt/tests/demo_project').ensuredir()
        self = ProjectStructure(dpath, mod_name='demopkg_mwe', use_src=False)
        self.generate()
        self.analyze()
        self.install()

    """
    def __init__(self, repo_dpath='.', mod_name='demopkg_mwe', use_src=True):
        import ubelt as ub
        self.root = ub.Path(repo_dpath)
        self.mod_name = mod_name
        self.use_src = use_src
        if use_src:
            self.python_relpath = ub.Path('src', 'python')
        else:
            self.python_relpath = ub.Path('.')
        self.cxx_relpath = ub.Path('src', 'cxx')
        self.cxx_path    = (self.root / self.cxx_relpath)
        self.python_path = (self.root / self.python_relpath)
        self.mod_dpath = (self.python_path / self.mod_name)

    def setup(self):
        self.generate()
        self.install()

    def teardown(self):
        self.uninstall()
        self.delete()

    def install(self):
        import sys
        import ubelt as ub
        ub.cmd([sys.executable, '-m', 'pip', 'install', '-e', self.root],
               verbose=3, check=True)

    def delete(self):
        self.root.delete()

    def uninstall(self):
        import sys
        import ubelt as ub
        ub.cmd([sys.executable, '-m', 'pip', 'uninstall', self.mod_name, '-y'],
               verbose=3, check=True)

    def generate(self, with_cxx=0):
        import ubelt as ub
        self.mod_dpath.delete().ensuredir()
        self.cxx_path.delete()
        (self.root / 'CMakeLists.txt').delete()
        (self.mod_dpath / '__init__.py').write_text('__version__ = "1.0.0"')

        if self.use_src:
            package_dir_line = ub.codeblock(
                f'''
                package_dir={{'': '{self.python_relpath}'}},
                ''')
        else:
            package_dir_line = ''

        # Give the MWE a CXX extension
        WITH_CXX = with_cxx
        if WITH_CXX:
            (self.root / 'pyproject.toml').write_text(ub.codeblock(
                '''
                [build-system]
                requires = ["setuptools>=41.0.1", "scikit-build>=0.11.1", "numpy", "ninja>=1.10.2", "cmake>=3.21.2", "cython>=0.29.24",]
                '''))
            (self.root / 'setup.py').write_text(ub.codeblock(
                f'''
                if __name__ == '__main__':
                    from skbuild import setup
                    from setuptools import find_packages
                    packages = find_packages('./{self.python_relpath}')
                    setup(
                        {package_dir_line}
                        install_requires=['packaging'],
                        name='{self.mod_name}',
                        version="1.0.0",
                        description='MWE of a binpy project',
                        packages=packages,
                        include_package_data=True,
                    )
                '''))
            self.cxx_path.ensuredir()
            (self.root / 'CMakeLists.txt').write_text(ub.codeblock(
                rf'''
                cmake_minimum_required(VERSION 3.13.0)
                project({self.mod_name} LANGUAGES C Fortran)

                find_package(PythonInterp REQUIRED)
                find_package(PythonLibs REQUIRED)

                ###
                # Private helper function to execute `python -c "<cmd>"`
                #
                # Runs a python command and populates an outvar with the result of stdout.
                # Be careful of indentation if `cmd` is multiline.
                #
                function(pycmd outvar cmd)
                  execute_process(
                    COMMAND "${{PYTHON_EXECUTABLE}}" -c "${{{{cmd}}}}"
                    RESULT_VARIABLE _exitcode
                    OUTPUT_VARIABLE _output)
                  if(NOT ${{_exitcode}} EQUAL 0)
                    message(ERROR "Failed when running python code: \"\"\"
                ${{cmd}}\"\"\"")
                    message(FATAL_ERROR "Python command failed with error code: ${{_exitcode}}")
                  endif()
                  # Remove supurflous newlines (artifacts of print)
                  string(STRIP "${{_output}}" _output)
                  set(${{outvar}} "${{_output}}" PARENT_SCOPE)
                endfunction()

                ###
                # Find scikit-build and include its cmake resource scripts
                #
                if (NOT SKBUILD)
                  pycmd(skbuild_location "import os, skbuild; print(os.path.dirname(skbuild.__file__))")
                  set(skbuild_cmake_dir "${{skbuild_location}}/resources/cmake")
                  # If skbuild is not the driver, then we need to include its utilities in our CMAKE_MODULE_PATH
                  list(APPEND CMAKE_MODULE_PATH ${{skbuild_cmake_dir}})
                endif()

                find_package(PythonExtensions REQUIRED)
                find_package(Cython REQUIRED)
                find_package(NumPy REQUIRED)

                # Backend C library
                add_subdirectory("src/cxx")

                # Cython library
                add_subdirectory("src/python/{self.mod_name}")
                '''))

            (self.cxx_path / 'myalgo.h').write_text(ub.codeblock(
                '''
                #ifndef MYALGO_H
                #define MYALGO_H
                int myalgo(long *arr1, long *arr2, size_t num);
                #endif MYALGO_H
                '''))
            (self.cxx_path / 'myalgo.c').write_text(ub.codeblock(
                r'''
                #include <string.h>
                long myalgo(long *arr1, long *arr2, size_t num)
                {
                    for (int i = 0 ; i < num ; i++ )
                    {
                        arr2[i] = arr1[i] + arr2[i];
                    }
                    return 1;
                }
                '''))
            cmake_list_cxx = self.cxx_path / 'CMakeLists.txt'
            cmake_list_cxx.write_text(ub.codeblock(
                '''
                set(MYALGO_MODULE_NAME "myalgo")
                list(APPEND MYALGO_SOURCES "myalgo.h" "myalgo.c")
                add_library(${MYALGO_MODULE_NAME} STATIC ${MYALGO_SOURCES})
                '''))

            (self.mod_dpath / 'myalgo_cython.pyx').write_text(ub.codeblock(
                '''
                import numpy as np
                cimport numpy as np
                cdef extern from "../../cxx/myalgo.h":
                    cdef int myalgo(long *arr1, long *arr2, size_t num);

                def call_myalgo():
                    """
                    This is a docstring
                    """
                    cdef int result;
                    cdef np.ndarray[np.int64_t, ndim=1] arr1
                    cdef np.ndarray[np.int64_t, ndim=1] arr2
                    arr1 = np.array([1, 2, 3], dtype=np.int64)
                    arr2 = np.array([4, 6, 9], dtype=np.int64)
                    cdef long [:] arr1_view = arr1
                    cdef long [:] arr2_view = arr2
                    cdef size_t num = len(arr1)
                    print(f'arr1={arr1}')
                    print(f'arr2={arr2}')
                    print('calling my algo')
                    result = myalgo(&arr1_view[0], &arr2_view[0], num)
                    print(f'arr1={arr1}')
                    print(f'arr2={arr2}')
                    return result
                '''))

            (self.mod_dpath / 'CMakeLists.txt').write_text(ub.codeblock(
                '''
                set(cython_source "myalgo_cython.pyx")
                set(PYMYALGO_MODULE_NAME "myalgo_cython")

                # Translate Cython into C/C++
                add_cython_target(${PYMYALGO_MODULE_NAME} "${cython_source}" C OUTPUT_VAR sources)

                # Add other C sources
                list(APPEND sources )

                # Create C++ library. Specify include dirs and link libs as normal
                add_library(${PYMYALGO_MODULE_NAME} MODULE ${sources})
                target_include_directories(
                    ${PYMYALGO_MODULE_NAME}
                    PUBLIC
                    ${NumPy_INCLUDE_DIRS}
                    ${PYTHON_INCLUDE_DIR}
                    ${CMAKE_CURRENT_SOURCE_DIR}
                )

                # TODO: not sure why this isn't set in the global scope?
                # Hack around it: just hard code the module name
                set(MYALGO_MODULE_NAME "myalgo")

                # TODO: linking to the MYALGO shared object isn't working 100% yet.
                target_link_libraries(${PYMYALGO_MODULE_NAME} ${MYALGO_MODULE_NAME})

                target_compile_definitions(${PYMYALGO_MODULE_NAME} PUBLIC
                    "NPY_NO_DEPRECATED_API"
                    #"NPY_1_7_API_VERSION=0x00000007"
                )

                # Transform the C++ library into an importable python module
                python_extension_module(${PYMYALGO_MODULE_NAME})

                # Install the C++ module to the correct relative location
                # (this will be an inplace build if you use `pip install -e`)
                #file(RELATIVE_PATH pymyalgo_install_dest "${CMAKE_SOURCE_DIR}" "${CMAKE_CURRENT_SOURCE_DIR}")

                # My "normal" method of setting install targets does not seem to work here. Hacking it.
                # NOTE: skbuild *seems* to place libraries in a data dir *unless* the install destination
                # corresponds exactly to the <package_dir>/<package_name> specified implicitly in setup.py
                set(pymyalgo_install_dest "src/python/{self.mod_name}")
                #install(TARGETS ${MYALGO_MODULE_NAME} LIBRARY DESTINATION "${pymyalgo_install_dest}")
                install(TARGETS ${PYMYALGO_MODULE_NAME} LIBRARY DESTINATION "${pymyalgo_install_dest}")
                '''
            ))
        else:
            # Pure Python
            # TODO: Might want to test with different build backends.
            (self.root / 'pyproject.toml').write_text(ub.codeblock(
                '''
                [build-system]
                requires = ["setuptools>=41.0.1", "wheel"]
                build-backend = "setuptools.build_meta"
                '''))
            (self.root / 'setup.py').write_text(ub.codeblock(
                f'''
                if __name__ == '__main__':
                    from setuptools import setup
                    from setuptools import find_packages
                    packages = find_packages('./{self.python_relpath}')
                    setup(
                        {package_dir_line}
                        package_data={{
                            '{self.mod_name}': ['py.typed', '*.pyi'],
                        }},
                        install_requires=['packaging'],
                        name='{self.mod_name}',
                        version="1.0.0",
                        description='MWE of a purepy project',
                        packages=packages,
                        include_package_data=True,
                    )
                '''))
            (self.mod_dpath / 'py.typed').write_text('')
            (self.mod_dpath / 'submod.py').write_text('A = 1')
            (self.mod_dpath / 'submod.pyi').write_text('A: int')

    def analyze(self):
        """
        For debugging and develoment only, don't run in the tests

        Requires:
            rich, xdev
        """
        from rich.console import Console
        from rich.panel import Panel
        from rich.syntax import Syntax
        from rich.table import Table
        import distutils.sysconfig
        import ubelt as ub
        import xdev

        console = Console()

        def rich_file_content(fpath, lexer='bash'):
            import os
            text = fpath.read_text()
            return Panel(Syntax(text, lexer), title=os.fspath(fpath))

        def print_egg_path_content(egg_info_dpath, color='blue'):
            blocklist = {'requires.txt'}
            fpaths = egg_info_dpath.ls()
            table = Table(f'[{color}]' + str(egg_info_dpath))
            for fpath in fpaths:
                if fpath.name not in blocklist:
                    panel = rich_file_content(fpath)
                    table.add_row(panel)
            console.print(table)

        print('\n')
        print('Repo Structure:')
        directory_blocklist = ['.*', '.git', 'dist', '_skbuild', 'dev']
        xdev.tree_repr(self.root, max_files=None, dirblocklist=directory_blocklist)

        seems_installed = 0

        print('\n')
        print('Content of the EGG Link:')
        site_dpath = ub.Path(distutils.sysconfig.get_python_lib())
        egg_link_fpaths = list(site_dpath.glob(self.mod_name.replace('_', '*') + '*.egg-link'))
        if len(egg_link_fpaths) == 0:
            console.print('[red] No egg link')
            seems_installed = 0
        else:
            assert len(egg_link_fpaths) == 1
            egg_link_fpath = egg_link_fpaths[0]
            console.print(rich_file_content(egg_link_fpath))
            seems_installed = 1

        # Note: (recently 2022-08-ish) python switched to a new type of
        # This is not present in setuptools==63.2.0 but is in 65.3.0
        # editable install. TODO: incomporate this.
        editable_fpaths = list(site_dpath.glob('__editable__*' + self.mod_name.replace('_', '*') + '*'))
        print(f'editable_fpaths={editable_fpaths}')

        print('\n')
        print('Check easy-install.pth')
        easy_install_fpath = site_dpath / 'easy-install.pth'
        assert easy_install_fpath.exists()
        easy_install_text = easy_install_fpath.read_text()
        abs_path = self.mod_dpath.absolute().parent
        print(f'abs_path={abs_path}')
        if str(abs_path)  in easy_install_text:
            console.print('[green] Easy install dpath is good')
        else:
            console.print('[red] Easy install does not contain this package')
            # console.print(rich_file_content(easy_install_fpath))

        expected_egg_info_dpath = self.python_path / f'{self.mod_name}.egg-info'
        all_egg_infos = [ub.Path(e).resolve() for e in xdev.find('*.egg-info', dpath=self.root, dirblocklist=directory_blocklist)]
        other_egg_infos = set(all_egg_infos) - {expected_egg_info_dpath.resolve()}
        print('expected_egg_info_dpath = {}'.format(ub.repr2(expected_egg_info_dpath, nl=1)))
        if expected_egg_info_dpath.exists():
            console.print('[green] Egg info exists in expected location')
            egg_info_dpath = expected_egg_info_dpath
            print_egg_path_content(egg_info_dpath, color='green')
        else:
            console.print('[red] Egg info does not exist in expected location')
            print(f'other_egg_infos={other_egg_infos}')

        if other_egg_infos:
            console.print('[red] THERE ARE UNEXEPCTED EGG INFOS')
            for egg_info_dpath in other_egg_infos:
                print_egg_path_content(egg_info_dpath, color='red')

        if seems_installed:
            print('\n')
            print('Test to ensure we can import the module')
            command = f'python -c "import {self.mod_name}; print({self.mod_name})"'
            info = ub.cmd(command, verbose=3)
            if info['ret'] != 0:
                raise Exception('failed to import')
            assert str(self.mod_dpath) in info['out']
        else:
            console.print('[yellow] Package does not seem installed, so skipping import test')

    def serialize_install(self):
        # TODO: serialize this step to make it fast
        import distutils.sysconfig
        import ubelt as ub
        site_dpath = ub.Path(distutils.sysconfig.get_python_lib())
        egg_link_fpaths = list(site_dpath.glob(self.mod_name.replace('_', '*') + '*.egg-link'))
        editable_fpaths = list(site_dpath.glob('__editable__*' + self.mod_name.replace('_', '*') + '*'))
        easy_install_fpath = site_dpath / 'easy-install.pth'  # NOQA
        print(f'egg_link_fpaths={egg_link_fpaths}')
        print(f'editable_fpaths={editable_fpaths}')


GLOBAL_PROJECTS = []


def _check_skip_editable_module_tests():
    UBELT_DO_EDITABLE_TESTS = os.environ.get('UBELT_DO_EDITABLE_TESTS', '')
    if not UBELT_DO_EDITABLE_TESTS:
        import pytest
        pytest.skip('UBELT_DO_EDITABLE_TESTS is not enabled')

    if sys.platform.startswith('win32'):
        import pytest
        pytest.skip('skip editable module tests on Win32')

    if sys.platform.startswith('freebsd'):
        import pytest
        pytest.skip('skip editable module tests on FreeBSD')


def setup_module(module):
    """ setup any state specific to the execution of the given module."""
    import uuid
    import ubelt as ub

    _check_skip_editable_module_tests()

    suffix = ub.hash_data(uuid.uuid4(), base='abc')[0:8]
    dpath = ub.Path.appdir('ubelt/tests/demo_packages').ensuredir()

    # Define pure python module with ./src/python structure
    mod_name = 'purepy_src_demo_pkg_' + suffix
    PUREPY_SRC_PROJECT = ProjectStructure(repo_dpath=dpath / mod_name,
                                          mod_name=mod_name, use_src=True)
    PUREPY_SRC_PROJECT.setup()
    GLOBAL_PROJECTS.append(PUREPY_SRC_PROJECT)

    if 0:
        self = PUREPY_SRC_PROJECT
        self.serialize()

    # Define pure python module with the package at root level
    mod_name = 'purepy_root_demo_pkg_' + suffix
    PUREPY_SRC_PROJECT = ProjectStructure(repo_dpath=dpath / mod_name,
                                          mod_name=mod_name, use_src=False)
    PUREPY_SRC_PROJECT.setup()
    GLOBAL_PROJECTS.append(PUREPY_SRC_PROJECT)

    if 0:
        for proj in GLOBAL_PROJECTS:
            proj.analyze()


def teardown_module(module):
    """ teardown any state that was previously setup with a setup_module
    method.
    """
    _check_skip_editable_module_tests()
    for PROJ in GLOBAL_PROJECTS:
        PROJ.teardown()


def test_import_of_editable_install():
    _check_skip_editable_module_tests()
    import ubelt as ub
    for PROJ in GLOBAL_PROJECTS:
        result = ub.modname_to_modpath(PROJ.mod_name)
        print(f'result={result}')
        assert result is not None
        assert PROJ.mod_dpath == ub.Path(result)



import ubelt as ub

DEBUG_PATH = 0
# ub.Path.home().name == 'joncrall'


def _demo_directory_structure():
    import ubelt as ub
    import uuid

    level = 0

    suffix = ub.hash_data(uuid.uuid4())[0:8]
    dpath = ub.Path.appdir('ubelt', 'tests', 'test_path')
    base = (dpath / suffix).delete().ensuredir()

    (base / 'root' / 'dir_L0_X0_A').ensuredir()

    if level > 2:
        (base / 'root' / 'dir_L0_X0_A' / 'dir_L1_X0_B').ensuredir()

    if level > 1:
        (base / 'root' / 'dir_L0_X1_C').ensuredir()

    (base / 'root' / 'inside_dir').ensuredir()
    (base / 'root' / 'links').ensuredir()
    (base / 'outside_dir').ensuredir()

    (base / 'root' / 'file_L0_X0_a.txt').touch()

    if level > 1:
        (base / 'root' / 'dir_L0_X0_A' / 'file_L1_X0_b.txt').touch()

    if level > 1:
        (base / 'root' / 'dir_L0_X1_C' / 'file_L1_X0_c.txt').touch()

    (base / 'root' / 'inside_dir' / 'inside_file.txt').touch()
    (base / 'outside_dir' / 'outside_file.txt').touch()

    # Create links inside and outside the root
    to_abs_symlink = []
    to_abs_symlink.append((base / 'root/inside_dir/inside_file.txt' , base / 'root/links/inside_flink.txt'))
    to_abs_symlink.append((base / 'outside_dir/outside_file.txt'    , base / 'root/links/outside_flink.txt'))
    to_abs_symlink.append((base / 'outside_dir'                     , base / 'root/links/outside_dlink'))
    to_abs_symlink.append((base / 'root/inside_dir'                 , base / 'root/links/inside_dlink'))
    to_abs_symlink.append((base / 'root/links/cyclic'                , (base / 'root/links/cyclic/n1/n2').ensuredir() / 'loop'))

    to_rel_symlink = []
    to_rel_symlink.append((base / 'root/inside_dir/inside_file.txt' , base / 'root/links/rel_inside_flink.txt'))
    to_rel_symlink.append((base / 'outside_dir/outside_file.txt'    , base / 'root/links/rel_outside_flink.txt'))
    to_rel_symlink.append((base / 'outside_dir'                     , base / 'root/links/rel_outside_dlink'))
    to_rel_symlink.append((base / 'root/inside_dir'                 , base / 'root/links/rel_inside_dlink'))
    to_rel_symlink.append((base / 'root/links/rel_cyclic'           , (base / 'root/links/rel_cyclic/n1/n2/').ensuredir() / 'rel_loop'))

    try:
        # TODO: the implementation of ubelt.symlink might be wrong when the
        # link target is relative.
        import os
        for real, link in to_abs_symlink:
            link.symlink_to(real)
            # ub.symlink(real, link, verbose=1)

        for real, link in to_rel_symlink:
            rel_real = os.path.relpath(real, link.parent)
            link.symlink_to(rel_real)
            # ub.symlink(rel_real, link, verbose=1)
    except Exception:
        import pytest
        pytest.skip('unable to symlink')

    if 0:
        import xdev
        xdev.tree_repr(base)
    return base


### MOVE TESTS


def test_move_dir_to_non_existing():
    base = _demo_directory_structure()
    root = base / 'root'

    if ub.LINUX:
        root2 = root.copy(root.augment(tail='2'))
        root3 = root.copy(root.augment(tail='3'))

    if DEBUG_PATH:
        import xdev
        xdev.tree_repr(base)

    root.move(base / 'our_move')

    if ub.LINUX:
        ub.cmd(f'mv {root2} {base}/linux_move', verbose=2, check=1)
        ub.cmd(f'mv -T {root3} {base}/linux_moveT', verbose=2, check=1)

    if DEBUG_PATH:
        import xdev
        xdev.tree_repr(base)

    if ub.LINUX:
        # We behave like Linux mv here in both cases here
        case1 = _comparable_walk(base / 'linux_move')
        case2 = _comparable_walk(base / 'linux_moveT')
        case3 = _comparable_walk(base / 'our_move')
        assert case1 == case2 == case3

    base.delete()


def test_move_to_nested_non_existing():
    base = _demo_directory_structure()
    root = base / 'root'

    if ub.LINUX:
        root2 = root.copy(root.augment(tail='2'))
        root3 = root.copy(root.augment(tail='3'))

    if DEBUG_PATH:
        import xdev
        xdev.tree_repr(base)

    # shutil move will make the parent directory if it doesn't exist.
    root.move(base / 'our/move')

    if ub.LINUX:
        # Posix fails unless the parent exists
        (base / 'linux').ensuredir()
        ub.cmd(f'mv -v {root2} {base}/linux/move', verbose=2, check=1)
        ub.cmd(f'mv -Tv {root3} {base}/linux/moveT', verbose=2, check=1)
    if DEBUG_PATH:
        import xdev
        xdev.tree_repr(base)

    if ub.LINUX:
        # We behave like Linux mv here in both cases here
        # up to the fact that we will always create the dir, whereas mv wont
        case1 = _comparable_walk(base / 'linux/move')
        case2 = _comparable_walk(base / 'linux/moveT')
        case3 = _comparable_walk(base / 'our/move')
        assert case1 == case2 == case3

    base.delete()


def test_move_dir_to_existing_dir_noconflict():
    base = _demo_directory_structure()
    root = base / 'root'

    (base / 'our_move').ensuredir()

    if ub.LINUX:
        root2 = root.copy(root.augment(tail='2'))
        root3 = root.copy(root.augment(tail='3'))
        (base / 'linux_move').ensuredir()
        (base / 'linux_moveT').ensuredir()

    if DEBUG_PATH:
        import xdev
        xdev.tree_repr(base)

    import pytest
    with pytest.raises(IOError):
        # shutil.move behaves similar to linux with -T
        # We are just going to disallow this case
        root.move(base / 'our_move')

    if ub.LINUX:
        ub.cmd(f'mv {root2} {base}/linux_move', verbose=2, check=1)
        ub.cmd(f'mv -T {root3} {base}/linux_moveT', verbose=2, check=1)

    if DEBUG_PATH:
        import xdev
        xdev.tree_repr(base)

    base.delete()


def test_move_dir_to_existing_dir_withconflict():
    base = _demo_directory_structure()

    root = base / 'root'
    bluntobject = (root / 'will_they_wont_they.txt')
    bluntobject.write_text('smash!')

    if ub.LINUX:
        root2 = root.copy(root.augment(tail='2'))
        root3 = root.copy(root.augment(tail='3'))  # NOQA

    dst1 = (base / 'our_move').ensuredir()
    dst2 = (base / 'linux_move').ensuredir()
    dst3 = (base / 'linux_move_T').ensuredir()

    toclobber1 = (dst1 / 'will_they_wont_they.txt')
    toclobber1.write_text('I hope nobody clobbers me!')
    disjoint1 = (dst1 / 'disjoint.txt')
    disjoint1.write_text('I should be disjoint!')

    toclobber2 = (dst2 / 'will_they_wont_they.txt')
    toclobber2.write_text('I hope nobody clobbers me!')
    disjoint2 = (dst2 / 'disjoint.txt')
    disjoint2.write_text('I should be disjoint!')

    toclobber3 = (dst3 / 'will_they_wont_they.txt')
    toclobber3.write_text('I hope nobody clobbers me!')
    disjoint3 = (dst3 / 'disjoint.txt')
    disjoint3.write_text('I should be disjoint!')

    if DEBUG_PATH:
        import xdev
        print('BEFORE MOVE')
        xdev.tree_repr(base)

    # This case is weird, dont let the user do it.
    # they can use shutil if they want
    import pytest
    with pytest.raises(IOError):
        root.move(dst1)

    if 0:
        if ub.LINUX:
            ub.cmd(f'mv -v {root2} {dst2}', verbose=2, check=1)
            # The mv command wont move a non-empty directory!
            # Maybe we shouldn't either.
            # ub.cmd(f'mv -T -u -f -v {root3} {dst3}', verbose=3, check=1)

        if DEBUG_PATH:
            import xdev
            print('AFTER MOVE')
            xdev.tree_repr(base)

        got = toclobber1.read_text()
        # THIS IS VERY SURPRISING, the file being moved is clobbered, but the file
        # in the dst is safe!
        assert got != 'smash!'
        assert not bluntobject.exists()

        if ub.LINUX:
            got2 = toclobber3.read_text()
            assert got2 == 'smash!'

        assert toclobber1.exists()
        assert bluntobject.exists()
        assert disjoint1.exists()
        assert disjoint1.read_text() == 'I should be disjoint!'

        if ub.LINUX:
            assert disjoint2.exists()
            assert disjoint2.read_text() == 'I should be disjoint!'
            assert disjoint3.exists()
            assert disjoint3.read_text() == 'I should be disjoint!'

    base.delete()

### Simple Copy Tests


def test_copy_basic():
    dpath = ub.Path.appdir('ubelt', 'tests', 'test_path', 'test_copy_basic')
    dpath.delete().ensuredir()
    fpath = (dpath / 'file.txt')
    fpath.write_text('foobar')
    empty_dpath = (dpath / 'empty_dir').ensuredir()
    full_dpath = (dpath / 'full_dir').ensuredir()
    (full_dpath / 'nested_file.txt').touch()

    if DEBUG_PATH:
        print('AFTER COPY')
        import xdev
        xdev.tree_repr(dpath)

    fpath.copy(fpath.augment(prefix='copied_'))
    empty_dpath.copy(empty_dpath.augment(prefix='copied_'))
    full_dpath.copy(full_dpath.augment(prefix='copied_'))

    # Doing it again will fail
    import pytest
    with pytest.raises(IOError):
        fpath.copy(fpath.augment(prefix='copied_'))
    with pytest.raises(IOError):
        empty_dpath.copy(empty_dpath.augment(prefix='copied_'))
    with pytest.raises(IOError):
        full_dpath.copy(full_dpath.augment(prefix='copied_'))

    # But with overwrite=True it is ok
    fpath.copy(fpath.augment(prefix='copied_'), overwrite=True)
    empty_dpath.copy(empty_dpath.augment(prefix='copied_'), overwrite=True)
    full_dpath.copy(full_dpath.augment(prefix='copied_'), overwrite=True)

    if DEBUG_PATH:
        print('AFTER COPY')
        import xdev
        xdev.tree_repr(dpath)


def test_copy_meta():
    dpath = ub.Path.appdir('ubelt', 'tests', 'test_path', 'test_copy_basic')
    dpath.delete().ensuredir()
    fpath = (dpath / 'file.txt')
    fpath.write_text('foobar')
    empty_dpath = (dpath / 'empty_dir').ensuredir()
    full_dpath = (dpath / 'full_dir').ensuredir()
    (full_dpath / 'nested_file.txt').touch()

    if DEBUG_PATH:
        print('AFTER COPY')
        import xdev
        xdev.tree_repr(dpath)

    for meta in ['stats', 'mode', None]:
        prefix = 'copied_' + str(meta) + '_'
        fpath.copy(fpath.augment(prefix=prefix), meta=meta)
        empty_dpath.copy(empty_dpath.augment(prefix=prefix))
        full_dpath.copy(full_dpath.augment(prefix=prefix))

    # TODO: verify that the metadata really did copy as intended
    if DEBUG_PATH:
        print('AFTER COPY')
        import xdev
        xdev.tree_repr(dpath)

### Simple Move Tests


def test_move_basic():
    dpath = ub.Path.appdir('ubelt', 'tests', 'test_path', 'test_move_basic')
    dpath.delete().ensuredir()
    fpath = (dpath / 'file.txt')
    fpath.write_text('foobar')
    empty_dpath = (dpath / 'empty_dir').ensuredir()
    full_dpath = (dpath / 'full_dir').ensuredir()
    (full_dpath / 'nested_file.txt').touch()

    if DEBUG_PATH:
        print('AFTER COPY')
        import xdev
        xdev.tree_repr(dpath)

    fpath.move(fpath.augment(prefix='moved_'))
    empty_dpath.move(empty_dpath.augment(prefix='moved_'))
    full_dpath.move(full_dpath.augment(prefix='moved_'))

    if DEBUG_PATH:
        print('AFTER COPY')
        import xdev
        xdev.tree_repr(dpath)


def test_move_meta():
    base_dpath = ub.Path.appdir('ubelt', 'tests', 'test_path', 'test_move_basic')
    base_dpath.delete().ensuredir()

    for meta in ['stats', 'mode', None]:
        prefix = 'copied_' + str(meta) + '_'
        dpath = (base_dpath / prefix).ensuredir()
        fpath = (dpath / 'file.txt')
        fpath.write_text('foobar')
        empty_dpath = (dpath / 'empty_dir').ensuredir()
        full_dpath = (dpath / 'full_dir').ensuredir()
        (full_dpath / 'nested_file.txt').touch()

        fpath.move(fpath.augment(prefix=prefix), meta=meta)
        empty_dpath.move(empty_dpath.augment(prefix=prefix))
        full_dpath.move(full_dpath.augment(prefix=prefix))

    # TODO: test that the metadata really did move as intended

    if DEBUG_PATH:
        print('AFTER MOVE')
        import xdev
        xdev.tree_repr(base_dpath)


### COPY TESTS


def test_copy_dir_to_non_existing():
    base = _demo_directory_structure()
    root = base / 'root'
    if DEBUG_PATH:
        print('BEFORE COPY')
        import xdev
        xdev.tree_repr(base)

    dst = root.copy(base / 'our_copy')

    if ub.LINUX:
        ub.cmd(f'cp -r {root} {base}/linux_copy', verbose=2)

    print(f'dst={dst}')
    if DEBUG_PATH:
        print('AFTER COPY')
        import xdev
        xdev.tree_repr(base)

    if ub.LINUX:
        # Our copy should behave like the linux copy
        case1 = _comparable_walk(base / 'our_copy')
        case2 = _comparable_walk(base / 'linux_copy')
        print('case1 = {}'.format(ub.urepr(case1, nl=1)))
        print('case2 = {}'.format(ub.urepr(case2, nl=1)))
        assert case1 == case2

    base.delete()


def test_copy_to_nested_non_existing_with_different_symlink_flags():
    base = _demo_directory_structure()
    root = base / 'root'
    if DEBUG_PATH:
        import xdev
        xdev.tree_repr(base)
    root.copy(base / 'new_subdir' / 'new_root_FD0_FF1', follow_dir_symlinks=False, follow_file_symlinks=True)
    root.copy(base / 'new_subdir' / 'new_root_FD0_FF0', follow_dir_symlinks=False, follow_file_symlinks=False)
    (root / 'links' / 'cyclic').delete()
    (root / 'links' / 'rel_cyclic').delete()
    root.copy(base / 'new_subdir' / 'new_root_FD1_FF1', follow_dir_symlinks=True, follow_file_symlinks=True)
    root.copy(base / 'new_subdir' / 'new_root_FD1_FF0', follow_dir_symlinks=True, follow_file_symlinks=False)
    if DEBUG_PATH:
        import xdev
        xdev.tree_repr(base)
    base.delete()


def test_copy_dir_to_existing_dir_noconflict():
    base = _demo_directory_structure()
    root = base / 'root'
    (root / 'links' / 'cyclic').delete()
    (root / 'links' / 'rel_cyclic').delete()

    dst1 = (base / 'our_copy').ensuredir()
    dst2 = (base / 'linux_copy').ensuredir()
    dst3 = (base / 'linux_copyT').ensuredir()

    if DEBUG_PATH:
        import xdev
        print('BEFORE MOVE')
        xdev.tree_repr(base)

    root.copy(dst1, overwrite=True)

    if ub.LINUX:
        # We behave like linux copy with T here.
        ub.cmd(f'cp -r {root} {dst2}', verbose=2)
        ub.cmd(f'cp -r -T {root} {dst3}', verbose=2)

    if DEBUG_PATH:
        import xdev
        print('AFTER MOVE')
        xdev.tree_repr(base)

    if ub.LINUX:
        # Our copy should behave like the linux copy
        case1 = _comparable_walk(base / 'our_copy')
        case2 = _comparable_walk(base / 'linux_copy')
        case3 = _comparable_walk(base / 'linux_copyT')
        assert case1 == case3
        assert case1 != case2
    base.delete()


def test_copy_dir_to_existing_dir_withconflict():
    base = _demo_directory_structure()

    root = base / 'root'
    bluntobject = (root / 'will_they_wont_they.txt')
    bluntobject.write_text('smash!')

    dst1 = (base / 'our_copy').ensuredir()
    dst2 = (base / 'linux_copy').ensuredir()
    dst3 = (base / 'linux_copyT').ensuredir()

    toclobber1 = (dst1 / 'will_they_wont_they.txt')
    toclobber1.write_text('I hope nobody clobbers me!')
    disjoint1 = (dst1 / 'disjoint.txt')
    disjoint1.write_text('I should be disjoint!')

    toclobber2 = (dst2 / 'will_they_wont_they.txt')
    toclobber2.write_text('I hope nobody clobbers me!')
    disjoint2 = (dst2 / 'disjoint.txt')
    disjoint2.write_text('I should be disjoint!')

    toclobber3 = (dst3 / 'will_they_wont_they.txt')
    toclobber3.write_text('I hope nobody clobbers me!')
    disjoint3 = (dst3 / 'disjoint.txt')
    disjoint3.write_text('I should be disjoint!')

    if DEBUG_PATH:
        import xdev
        print('BEFORE MOVE')
        xdev.tree_repr(base)

    root.copy(dst1, overwrite=True)

    if ub.LINUX:
        ub.cmd(f'cp -r {root} {dst2}', verbose=2, check=1)
        ub.cmd(f'cp -r -T {root} {dst3}', verbose=2, check=1)

    if DEBUG_PATH:
        import xdev
        print('AFTER MOVE')
        xdev.tree_repr(base)

    # This behavior makes more sense to me
    got = toclobber1.read_text()
    assert got == 'smash!'

    if ub.LINUX:
        got2 = toclobber3.read_text()
        assert got2 == 'smash!'

    assert toclobber1.exists()
    assert bluntobject.exists()
    assert disjoint1.exists()
    assert disjoint1.read_text() == 'I should be disjoint!'

    if ub.LINUX:
        assert disjoint2.exists()
        assert disjoint2.read_text() == 'I should be disjoint!'
        assert disjoint3.exists()
        assert disjoint3.read_text() == 'I should be disjoint!'

    if ub.LINUX:
        # Our copy should behave like the linux copy
        case1 = _comparable_walk(base / 'our_copy')
        case2 = _comparable_walk(base / 'linux_copy')
        case3 = _comparable_walk(base / 'linux_copyT')
        print('case1 = {}'.format(ub.urepr(case1, nl=1)))
        print('case3 = {}'.format(ub.urepr(case3, nl=1)))
        print('case2 = {}'.format(ub.urepr(case2, nl=1)))
        assert case1 == case3
        assert case1 != case2
    base.delete()


def _comparable_walk(p):
    return sorted([(tuple(sorted(f)), tuple(sorted(d))) for (r, f, d) in (p).walk()])



import ubelt as ub
import pytest


def test_auto_dict():
    auto = ub.AutoDict()
    assert 0 not in auto
    auto[0][10][100] = None
    assert 0 in auto
    assert isinstance(auto[0], ub.AutoDict)


def test_auto_dict_to_dict():
    from ubelt.util_dict import AutoDict
    auto = AutoDict()
    auto[1] = 1
    auto['n1'] = AutoDict()
    auto['n1']['n2'] = AutoDict()
    auto['n1']['n2'][2] = 2
    auto['n1']['n2'][3] = 3
    auto['dict'] = {}
    auto['dict']['n3'] = AutoDict()
    auto['dict']['n3']['n4'] = AutoDict()
    print('auto = {!r}'.format(auto))
    static = auto.to_dict()
    print('static = {!r}'.format(static))
    assert not isinstance(static, AutoDict), '{}'.format(type(static))
    assert not isinstance(static['n1'], AutoDict), '{}'.format(type(static['n1']))
    assert not isinstance(static['n1']['n2'], AutoDict)
    assert isinstance(static['dict']['n3'], AutoDict)
    assert isinstance(static['dict']['n3']['n4'], AutoDict)


def test_auto_dict_ordered():
    # To Dict should repsect ordering
    from ubelt.util_dict import AutoOrderedDict, AutoDict
    auto = AutoOrderedDict()
    auto[0][3] = 3
    auto[0][2] = 2
    auto[0][1] = 1
    auto[0][4] = AutoDict()
    assert isinstance(auto, AutoDict)
    print('auto = {!r}'.format(auto))
    static = auto.to_dict()
    print('static = {!r}'.format(static))
    assert not isinstance(static, AutoDict), 'bad cast {}'.format(type(static))
    assert not isinstance(static[0][4], AutoDict), 'bad cast {}'.format(type(static[0][4]))
    assert list(auto[0].values())[0:3] == [3, 2, 1], 'maintain order'


def test_dzip_errors():
    with pytest.raises(TypeError):
        ub.dzip([1], 2)
    with pytest.raises(TypeError):
        ub.dzip(1, [2])
    with pytest.raises(ValueError):
        ub.dzip([1, 2, 3], [])
    with pytest.raises(ValueError):
        ub.dzip([], [4, 5, 6])
    with pytest.raises(ValueError):
        ub.dzip([1, 2, 3], [4, 5])


def test_group_items_callable():
    pairs = [
        ('ham', 'protein'),
        ('jam', 'fruit'),
        ('spam', 'protein'),
        ('eggs', 'protein'),
        ('cheese', 'dairy'),
        ('banana', 'fruit'),
    ]
    items, groupids = zip(*pairs)
    lut = dict(zip(items, groupids))

    result1 = ub.group_items(items, groupids)
    result2 = ub.group_items(items, lut.__getitem__)

    result1 = ub.map_values(set, result1)
    result2 = ub.map_values(set, result2)
    assert result1 == result2


def test_dict_hist_ordered():
    import random
    import string
    import ubelt as ub
    rng = random.Random(0)
    items = [rng.choice(string.ascii_letters) for _ in range(100)]
    # Ensure that the ordered=True bug is fixed
    a = sorted(ub.dict_hist(items, ordered=True).items())
    b = sorted(ub.dict_hist(items, ordered=False).items())
    assert a == b


def test_dict_subset_iterable():
    """
    CommandLine:
        xdoctest -m ~/code/ubelt/tests/test_dict.py test_dict_subset_iterable
    """
    # There was a bug in 0.7.0 where iterable keys would be exhausted too soon
    keys_list = list(range(10))
    dict_ = {k: k for k in keys_list}
    got = ub.dict_subset(dict_, iter(keys_list))
    assert dict(got) == dict_


# def _benchmark_groupid_sorted():
#     import random
#     import ubelt as ub

#     ydata = ub.ddict(list)
#     xdata = []

#     ti = ub.Timerit(100, bestof=10, verbose=True)

#     num = 10
#     for gamma in [0.01, .1, .5]:
#         for num in [10, 100, 1000, 10000, 100000]:
#             items = [random.random() for _ in range(num)]
#             groupids = [random.randint(0, int(num ** gamma)) for _ in range(num)]

#             xdata.append(num)

#             for timer in ti.reset(label='sort_g{}'.format(gamma)):
#                 with timer:
#                     ub.group_items(items, groupids, sorted_=True)
#             ydata[ti.label].append(ti.min())

#             for timer in ti.reset(label='nosort_g{}'.format(gamma)):
#                 with timer:
#                     ub.group_items(items, groupids, sorted_=False)
#             ydata[ti.label].append(ti.min())

#     ydata = ub.odict(sorted(ydata.items(), key=lambda t: t[1][-1])[::-1])
#     import netharn as nh
#     nh.util.autompl()
#     nh.util.multi_plot(xdata, ydata)



from __future__ import unicode_literals
from os.path import os


def test_touch():
    import ubelt as ub
    dpath = ub.Path.appdir('ubelt', 'tests').ensuredir()
    fpath = dpath / 'touch_file'
    assert not fpath.exists()
    ub.touch(fpath, verbose=True)
    assert fpath.exists()
    os.unlink(fpath)


def test_readwrite():
    import pytest
    import ubelt as ub
    dpath = ub.Path.appdir('ubelt', 'tests').ensuredir()
    fpath = dpath / 'testwrite.txt'
    if fpath.exists():
        os.remove(fpath)
    to_write = 'utf-8 symbols Δ, Й, ק, م, ๗, あ, 叶, 葉, and 말.'
    with pytest.warns(DeprecationWarning):
        ub.writeto(fpath, to_write, verbose=True)
    with pytest.warns(DeprecationWarning):
        read_ = ub.readfrom(fpath, verbose=True)
    assert read_ == to_write




def test_compatible_keywords():
    import ubelt as ub
    def func(a, e, f, *args, **kwargs):
        return a * e * f

    config = {
        'a': 2, 'b': 3, 'c': 7,
        'd': 11, 'e': 13, 'f': 17,
    }

    assert ub.compatible(config, func, keywords=True) is config
    assert ub.compatible(config, func, keywords=1) is config
    assert ub.compatible(config, func, keywords='truthy') is config

    assert ub.compatible(config, func, keywords=['iterable']) is not config
    assert ub.compatible(config, func, keywords=0) is not config
    assert ub.compatible(config, func, keywords={'b'}) == {'a': 2, 'e': 13, 'f': 17, 'b': 3}


def test_positional_only_args():
    import ubelt as ub
    import sys
    import pytest
    if sys.version_info[0:2] <= (3, 7):
        pytest.skip('position only arguments syntax requires Python >= 3.8')

    # Define via an exec, so this test does not raise a syntax error
    # in other versions of python and skips gracefully
    pos_only_code = ub.codeblock(
        '''
        import ubelt as ub
        def func(a, e, /,  f):
            return a * e * f
        ''')
    ns = {}
    exec(pos_only_code, ns, ns)
    func = ns['func']
    config = {
        'a': 2, 'b': 3, 'c': 7,
        'd': 11, 'e': 13, 'f': 17,
    }
    pos_only = ub.compatible(config, func)
    assert sorted(pos_only) == ['f']
